from typing import List, Optional, Union

import numpy as np
import torch
from astropy.io import fits

from .image_object import Image, ImageList
from .window import Window
from .jacobian_image import JacobianImage, JacobianImageList
from .model_image import ModelImage, ModelImageList
from .psf_image import PSFImage
from .. import config
from ..errors import InvalidImage
from .mixins import DataMixin

__all__ = ["TargetImage", "TargetImageList"]


class TargetImage(DataMixin, Image):
    """Image object which represents the data to be fit by a model. It can
    include a variance image, mask, and PSF as anciliary data which
    describes the target image.

    Target images are a basic unit of data in `AstroPhot`, they store
    the information collected from telescopes for which models are to
    be fit. There is minimal functionality in the `Target_Image`
    object itself, it is mostly defined in terms of how other objects
    interact with it.

    Basic usage:

    .. code-block:: python

      import astrophot as ap

      # Create target image
      image = ap.image.Target_Image(
          data="pixel data",
          wcs="astropy WCS object",
          variance="pixel uncertainties",
          psf="point spread function as PSF_Image object",
          mask=" True for pixels to ignore",
      )

      # Display the data
      fig, ax = plt.subplots()
      ap.plots.target_image(fig, ax, image)
      plt.show()

      # Save the image
      image.save("mytarget.fits")

      # Load the image
      image2 = ap.image.Target_Image(filename="mytarget.fits")

      # Make low resolution version
      lowrez = image.reduce(2)

    Some important information to keep in mind. First, providing an
    `astropy WCS` object is the best way to keep track of coordinates
    and pixel scale properties, especially when dealing with
    multi-band data. If images have relative positioning, rotation,
    pixel sizes, field of view this will all be handled automatically
    by taking advantage of `WCS` objects. Second, Providing accurate
    variance (or weight) maps is critical to getting a good fit to the
    data. This is a very common source of issues so it is worthwhile
    to review literature on how best to construct such a map. A good
    starting place is the FAQ for GALFIT:
    https://users.obs.carnegiescience.edu/peng/work/galfit/CHI2.html
    which is an excellent resource for all things image modeling. Just
    note that `AstroPhot` uses variance or weight maps, not sigma
    images. `AstroPhot` will not crete a variance map for the user, by
    default it will just assume uniform variance which is rarely
    accurate. Third, The PSF pixelscale must be a multiple of the
    image pixelscale. So if the image has a pixelscale of 1 then the
    PSF must have a pixelscale of 1, 1/2, 1/3, etc for anything to
    work out. Note that if the PSF pixelscale is finer than the image,
    then all modelling will be done at the higher resolution. This is
    recommended for accuracy though it can mean higher memory
    consumption.

    """

    def __init__(self, *args, psf=None, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.has_psf:
            self.psf = psf

    @property
    def has_psf(self):
        """Returns True when the target image object has a PSF model."""
        try:
            return self._psf is not None
        except AttributeError:
            return False

    @property
    def psf(self):
        """The PSF for the `Target_Image`. This is used to convolve the
        model with the PSF before evaluating the likelihood. The PSF
        should be a `PSF_Image` object or an `AstroPhot` PSF_Model.

        If no PSF is provided, then the image will not be convolved
        with a PSF and the model will be evaluated directly on the
        image pixels.

        """
        try:
            return self._psf
        except AttributeError:
            return None

    @psf.setter
    def psf(self, psf):
        """Provide a psf for the `Target_Image`. This is stored and passed to
        models which need to be convolved.

        The PSF doesn't need to have the same pixelscale as the
        image. It should be some multiple of the resolution of the
        `Target_Image` though. So if the image has a pixelscale of 1,
        the psf may have a pixelscale of 1, 1/2, 1/3, 1/4 and so on.

        """
        if hasattr(self, "_psf"):
            del self._psf  # remove old psf if it exists
        from ..models import Model

        if psf is None:
            self._psf = None
        elif isinstance(psf, PSFImage):
            self._psf = psf
        elif isinstance(psf, Model):
            self._psf = psf
        else:
            self._psf = PSFImage(
                data=psf,
                CD=self.CD,
                name=self.name + "_psf",
            )

    def copy(self, **kwargs):
        """Produce a copy of this image with all of the same properties. This
        can be used when one wishes to make temporary modifications to
        an image and then will want the original again.

        """
        kwargs = {"psf": self.psf, **kwargs}
        return super().copy(**kwargs)

    def fits_images(self):
        images = super().fits_images()
        if self.has_psf:
            if isinstance(self.psf, PSFImage):
                images.append(
                    fits.ImageHDU(
                        torch.transpose(self.psf.data, 0, 1).detach().cpu().numpy(),
                        name="PSF",
                        header=fits.Header(self.psf.fits_info()),
                    )
                )
            else:
                config.logger.warning("Unable to save PSF to FITS, not a PSF_Image.")
        return images

    def load(self, filename: str, hduext=0):
        """Load the image from a FITS file. This will load the data, WCS, and
        any ancillary data such as variance, mask, and PSF.

        """
        hdulist = super().load(filename, hduext=hduext)
        if "PSF" in hdulist:
            self.psf = PSFImage(
                data=np.array(hdulist["PSF"].data, dtype=np.float64),
                CD=(
                    (hdulist["PSF"].header["CD1_1"], hdulist["PSF"].header["CD1_2"]),
                    (hdulist["PSF"].header["CD2_1"], hdulist["PSF"].header["CD2_2"]),
                ),
            )
        return hdulist

    def jacobian_image(
        self,
        parameters: List[str],
        data: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        """
        Construct a blank `Jacobian_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        if data is None:
            data = torch.zeros(
                (*self.data.shape, len(parameters)),
                dtype=config.DTYPE,
                device=config.DEVICE,
            )
        kwargs = {
            "CD": self.CD.value,
            "crpix": self.crpix,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name + "_jacobian",
            **kwargs,
        }
        return JacobianImage(parameters=parameters, _data=data, **kwargs)

    def model_image(self, upsample=1, pad=0, **kwargs):
        """
        Construct a blank `Model_Image` object formatted like this current `Target_Image` object. Mostly used internally.
        """
        kwargs = {
            "_data": torch.zeros(
                (self.data.shape[0] * upsample + 2 * pad, self.data.shape[1] * upsample + 2 * pad),
                dtype=self.data.dtype,
                device=self.data.device,
            ),
            "CD": self.CD.value / upsample,
            "crpix": (self.crpix + 0.5) * upsample + pad - 0.5,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name + "_model",
            **kwargs,
        }
        return ModelImage(**kwargs)

    def psf_image(self, data, upscale=1, **kwargs):
        kwargs = {
            "data": data,
            "CD": self.CD.value / upscale,
            "identity": self.identity,
            "name": self.name + "_psf",
            **kwargs,
        }
        return PSFImage(**kwargs)

    def reduce(self, scale, **kwargs):
        """Returns a new `Target_Image` object with a reduced resolution
        compared to the current image. `scale` should be an integer
        indicating how much to reduce the resolution. If the
        `Target_Image` was originally (48,48) pixels across with a
        pixelscale of 1 and `reduce(2)` is called then the image will
        be (24,24) pixels and the pixelscale will be 2. If `reduce(3)`
        is called then the returned image will be (16,16) pixels
        across and the pixelscale will be 3.

        """

        return super().reduce(
            scale=scale,
            psf=(self.psf.reduce(scale) if isinstance(self.psf, PSFImage) else None),
            **kwargs,
        )


class TargetImageList(ImageList):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not all(isinstance(image, (TargetImage, TargetImageList)) for image in self.images):
            raise InvalidImage(
                f"Target_Image_List can only hold Target_Image objects, not {tuple(type(image) for image in self.images)}"
            )

    @property
    def variance(self):
        return tuple(image.variance for image in self.images)

    @variance.setter
    def variance(self, variance):
        for image, var in zip(self.images, variance):
            image.variance = var

    @property
    def has_variance(self):
        return any(image.has_variance for image in self.images)

    @property
    def weight(self):
        return tuple(image.weight for image in self.images)

    @weight.setter
    def weight(self, weight):
        for image, wgt in zip(self.images, weight):
            image.weight = wgt

    @property
    def has_weight(self):
        return any(image.has_weight for image in self.images)

    def jacobian_image(self, parameters: List[str], data: Optional[List[torch.Tensor]] = None):
        if data is None:
            data = tuple(None for _ in range(len(self.images)))
        return JacobianImageList(
            list(image.jacobian_image(parameters, dat) for image, dat in zip(self.images, data))
        )

    def model_image(self):
        return ModelImageList(list(image.model_image() for image in self.images))

    @property
    def mask(self):
        return tuple(image.mask for image in self.images)

    @mask.setter
    def mask(self, mask):
        for image, M in zip(self.images, mask):
            image.mask = M

    @property
    def has_mask(self):
        return any(image.has_mask for image in self.images)

    @property
    def psf(self):
        return tuple(image.psf for image in self.images)

    @psf.setter
    def psf(self, psf):
        for image, P in zip(self.images, psf):
            image.psf = P

    @property
    def has_psf(self):
        return any(image.has_psf for image in self.images)



from typing import Sequence, Literal

import torch
from scipy.optimize import minimize

from .base import BaseOptimizer
from .. import config
from ..errors import OptimizeStopSuccess

__all__ = ("ScipyFit",)


class ScipyFit(BaseOptimizer):

    def __init__(
        self,
        model,
        initial_state: Sequence = None,
        method: Literal[
            "Nelder-Mead", "L-BFGS-B", "TNC", "SLSQP", "Powell", "trust-constr"
        ] = "Nelder-Mead",
        ndf=None,
        **kwargs,
    ):

        super().__init__(model, initial_state, **kwargs)
        self.method = method
        # mask
        fit_mask = self.model.fit_mask()
        if isinstance(fit_mask, tuple):
            fit_mask = torch.cat(tuple(FM.flatten() for FM in fit_mask))
        else:
            fit_mask = fit_mask.flatten()
        if torch.sum(fit_mask).item() == 0:
            fit_mask = None

        if model.target.has_mask:
            mask = self.model.target[self.fit_window].flatten("mask")
            if fit_mask is not None:
                mask = mask | fit_mask
            self.mask = ~mask
        elif fit_mask is not None:
            self.mask = ~fit_mask
        else:
            self.mask = torch.ones_like(
                self.model.target[self.fit_window].flatten("data"), dtype=torch.bool
            )
        if self.mask is not None and torch.sum(self.mask).item() == 0:
            raise OptimizeStopSuccess("No data to fit. All pixels are masked")

        # Initialize optimizer attributes
        self.Y = self.model.target[self.fit_window].flatten("data")[self.mask]

        # 1 / (sigma^2)
        kW = kwargs.get("W", None)
        if kW is not None:
            self.W = torch.as_tensor(kW, dtype=config.DTYPE, device=config.DEVICE).flatten()[
                self.mask
            ]
        elif model.target.has_variance:
            self.W = self.model.target[self.fit_window].flatten("weight")[self.mask]
        else:
            self.W = torch.ones_like(self.Y)

        # The forward model which computes the output image given input parameters
        self.forward = lambda x: model(window=self.fit_window, params=x).flatten("data")[self.mask]
        # Compute the jacobian in representation units (defined for -inf, inf)
        self.jacobian = lambda x: model.jacobian(window=self.fit_window, params=x).flatten("data")[
            self.mask
        ]

        # variable to store covariance matrix if it is ever computed
        self._covariance_matrix = None

        # Degrees of freedom
        if ndf is None:
            self.ndf = max(1.0, len(self.Y) - len(self.current_state))
        else:
            self.ndf = ndf

    def chi2_ndf(self, x):
        return torch.sum(self.W * (self.Y - self.forward(x)) ** 2) / self.ndf

    def numpy_bounds(self):
        """Convert the model's parameter bounds to a format suitable for scipy.optimize."""
        bounds = []
        for param in self.model.dynamic_params:
            if param.shape == ():
                bound = [None, None]
                if param.valid[0] is not None:
                    bound[0] = param.valid[0].detach().cpu().numpy()
                if param.valid[1] is not None:
                    bound[1] = param.valid[1].detach().cpu().numpy()
                bounds.append(tuple(bound))
            else:
                for i in range(param.value.numel()):
                    bound = [None, None]
                    if param.valid[0] is not None:
                        bound[0] = param.valid[0].flatten()[i].detach().cpu().numpy()
                    if param.valid[1] is not None:
                        bound[1] = param.valid[1].flatten()[i].detach().cpu().numpy()
                    bounds.append(tuple(bound))
        return bounds

    def fit(self):

        res = minimize(
            lambda x: self.chi2_ndf(
                torch.tensor(x, dtype=config.DTYPE, device=config.DEVICE)
            ).item(),
            self.current_state,
            method=self.method,
            bounds=self.numpy_bounds(),
            options={
                "maxiter": self.max_iter,
            },
        )
        self.scipy_res = res
        self.message = self.message + f"success: {res.success}, message: {res.message}"
        self.current_state = torch.tensor(res.x, dtype=config.DTYPE, device=config.DEVICE)
        if self.verbose > 0:
            config.logger.info(
                f"Final Chi^2/DoF: {self.chi2_ndf(self.current_state):.6g}. Converged: {self.message}"
            )
        self.model.fill_dynamic_values(self.current_state)

        return self



import torch
import numpy as np


def default_prof(shape, pixelscale, min_pixels=2, scale=0.2):
    prof = [0, min_pixels * pixelscale]
    while prof[-1] < (np.max(shape) * pixelscale / 2):
        prof.append(prof[-1] + max(min_pixels * pixelscale, prof[-1] * scale))
    return np.array(prof)


def interp1d_torch(x_in, y_in, x_out):
    indices = torch.searchsorted(x_in[:-1], x_out) - 1
    weights = (y_in[1:] - y_in[:-1]) / (x_in[1:] - x_in[:-1])
    return y_in[indices] + weights[indices] * (x_out - x_in[indices])


def interp2d(
    im: torch.Tensor,
    x: torch.Tensor,
    y: torch.Tensor,
    padding_mode: str = "zeros",
) -> torch.Tensor:
    """
    Interpolates a 2D image at specified coordinates.
    Similar to `torch.nn.functional.grid_sample` with `align_corners=False`.

    Args:
        im (Tensor): A 2D tensor representing the image.
        x (Tensor): A tensor of x coordinates (in pixel space) at which to interpolate.
        y (Tensor): A tensor of y coordinates (in pixel space) at which to interpolate.

    Returns:
        Tensor: Tensor with the same shape as `x` and `y` containing the interpolated values.
    """

    # Convert coordinates to pixel indices
    h, w = im.shape

    # reshape for indexing purposes
    start_shape = x.shape
    x = x.flatten()
    y = y.flatten()

    if padding_mode == "zeros":
        valid = (x >= -0.5) & (x <= (w - 0.5)) & (y >= -0.5) & (y <= (h - 0.5))
    elif padding_mode == "border":
        x = x.clamp(-0.5, w - 0.5)
        y = y.clamp(-0.5, h - 0.5)
    else:
        raise ValueError(f"Unsupported padding mode: {padding_mode}")

    x0 = x.floor().long()
    y0 = y.floor().long()
    x0 = x0.clamp(0, w - 2)
    x1 = x0 + 1
    y0 = y0.clamp(0, h - 2)
    y1 = y0 + 1

    fa = im[y0, x0]
    fb = im[y1, x0]
    fc = im[y0, x1]
    fd = im[y1, x1]

    wa = (x1 - x) * (y1 - y)
    wb = (x1 - x) * (y - y0)
    wc = (x - x0) * (y1 - y)
    wd = (x - x0) * (y - y0)

    result = fa * wa + fb * wb + fc * wc + fd * wd

    if padding_mode == "zeros":
        return (result * valid).reshape(start_shape)
    elif padding_mode == "border":
        return result.reshape(start_shape)


def interp2d_ij(
    im: torch.Tensor,
    i: torch.Tensor,
    j: torch.Tensor,
) -> torch.Tensor:
    """
    Interpolates a 2D image at specified coordinates.
    Similar to `torch.nn.functional.grid_sample` with `align_corners=False`.

    Args:
        im (Tensor): A 2D tensor representing the image.
        x (Tensor): A tensor of x coordinates (in pixel space) at which to interpolate.
        y (Tensor): A tensor of y coordinates (in pixel space) at which to interpolate.

    Returns:
        Tensor: Tensor with the same shape as `x` and `y` containing the interpolated values.
    """

    # Convert coordinates to pixel indices
    h, w = im.shape

    # reshape for indexing purposes
    start_shape = i.shape
    i = i.flatten()
    j = j.flatten()

    # valid
    valid = (i >= -0.5) & (i <= (h - 0.5)) & (j >= -0.5) & (j <= (w - 0.5))

    i0 = i.floor().long()
    j0 = j.floor().long()
    i0 = i0.clamp(0, h - 2)
    i1 = i0 + 1
    j0 = j0.clamp(0, w - 2)
    j1 = j0 + 1

    fa = im[i0, j0]
    fb = im[i0, j1]
    fc = im[i1, j0]
    fd = im[i1, j1]

    wa = (i1 - i) * (j1 - j)
    wb = (i1 - i) * (j - j0)
    wc = (i - i0) * (j1 - j)
    wd = (i - i0) * (j - j0)

    result = fa * wa + fb * wb + fc * wc + fd * wd

    return (result * valid).view(*start_shape)



from .galaxy_model_object import GalaxyModel
from .psf_model_object import PSFModel
from .mixins import (
    NukerMixin,
    RadialMixin,
    iNukerMixin,
    RayMixin,
    WedgeMixin,
    SuperEllipseMixin,
    FourierEllipseMixin,
    WarpMixin,
)
from ..utils.decorators import combine_docstrings


__all__ = [
    "NukerGalaxy",
    "NukerPSF",
    "NukerSuperEllipse",
    "NukerFourierEllipse",
    "NukerWarp",
    "NukerWedge",
    "NukerRay",
]


@combine_docstrings
class NukerGalaxy(NukerMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class NukerPSF(NukerMixin, RadialMixin, PSFModel):
    _parameter_specs = {"Ib": {"units": "flux/arcsec^2", "value": 1.0}}
    usable = True


@combine_docstrings
class NukerSuperEllipse(NukerMixin, SuperEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class NukerFourierEllipse(NukerMixin, FourierEllipseMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class NukerWarp(NukerMixin, WarpMixin, RadialMixin, GalaxyModel):
    usable = True


@combine_docstrings
class NukerRay(iNukerMixin, RayMixin, GalaxyModel):
    usable = True


@combine_docstrings
class NukerWedge(iNukerMixin, WedgeMixin, GalaxyModel):
    usable = True



import torch

from .target_image import TargetImage
from .model_image import ModelImage
from .mixins import SIPMixin


class SIPModelImage(SIPMixin, ModelImage):

    def crop(self, pixels, **kwargs):
        """
        Crop the image by the number of pixels given. This will crop
        the image in all four directions by the number of pixels given.
        """
        if isinstance(pixels, int):  # same crop in all dimension
            crop = (slice(pixels, -pixels), slice(pixels, -pixels))
        elif len(pixels) == 1:  # same crop in all dimension
            crop = (slice(pixels[0], -pixels[0]), slice(pixels[0], -pixels[0]))
        elif len(pixels) == 2:  # different crop in each dimension
            crop = (
                slice(pixels[1], -pixels[1]),
                slice(pixels[0], -pixels[0]),
            )
        elif len(pixels) == 4:  # different crop on all sides
            crop = (
                slice(pixels[0], -pixels[1]),
                slice(pixels[2], -pixels[3]),
            )
        else:
            raise ValueError(
                f"Invalid crop shape {pixels}, must be int, (int,), (int, int), or (int, int, int, int)!"
            )
        kwargs = {
            "pixel_area_map": self.pixel_area_map[crop],
            "distortion_ij": self.distortion_ij[:, crop[0], crop[1]],
            "distortion_IJ": self.distortion_IJ[:, crop[0], crop[1]],
            **kwargs,
        }
        return super().crop(pixels, **kwargs)

    def reduce(self, scale: int, **kwargs):
        """This operation will downsample an image by the factor given. If
        scale = 2 then 2x2 blocks of pixels will be summed together to
        form individual larger pixels. A new image object will be
        returned with the appropriate pixelscale and data tensor. Note
        that the window does not change in this operation since the
        pixels are condensed, but the pixel size is increased
        correspondingly.

        Parameters:
            scale: factor by which to condense the image pixels. Each scale X scale region will be summed [int]

        """
        if not isinstance(scale, int) and not (
            isinstance(scale, torch.Tensor) and scale.dtype is torch.int32
        ):
            raise SpecificationConflict(f"Reduce scale must be an integer! not {type(scale)}")
        if scale == 1:
            return self

        MS = self.data.shape[0] // scale
        NS = self.data.shape[1] // scale

        kwargs = {
            "pixel_area_map": (
                self.pixel_area_map[: MS * scale, : NS * scale]
                .reshape(MS, scale, NS, scale)
                .sum(axis=(1, 3))
            ),
            "distortion_ij": (
                self.distortion_ij[:, : MS * scale, : NS * scale]
                .reshape(2, MS, scale, NS, scale)
                .mean(axis=(2, 4))
            ),
            "distortion_IJ": (
                self.distortion_IJ[:, : MS * scale, : NS * scale]
                .reshape(2, MS, scale, NS, scale)
                .mean(axis=(2, 4))
            ),
            **kwargs,
        }
        return super().reduce(
            scale=scale,
            **kwargs,
        )

    def fluxdensity_to_flux(self):
        self._data = self.data * self.pixel_area_map


class SIPTargetImage(SIPMixin, TargetImage):
    """
    A TargetImage with SIP distortion coefficients.
    This class is used to represent a target image with SIP distortion coefficients.
    It inherits from TargetImage and SIPMixin.
    """

    def model_image(self, upsample=1, pad=0, **kwargs):
        new_area_map = self.pixel_area_map
        new_distortion_ij = self.distortion_ij
        new_distortion_IJ = self.distortion_IJ
        if upsample > 1:
            U = torch.nn.Upsample(scale_factor=upsample, mode="nearest")
            new_area_map = (
                U(new_area_map.unsqueeze(0).unsqueeze(0)).squeeze(0).squeeze(0) / upsample**2
            )
            U = torch.nn.Upsample(scale_factor=upsample, mode="bilinear", align_corners=False)
            new_distortion_ij = U(self.distortion_ij.unsqueeze(1)).squeeze(1)
            new_distortion_IJ = U(self.distortion_IJ.unsqueeze(1)).squeeze(1)
        if pad > 0:
            new_area_map = (
                torch.nn.functional.pad(
                    new_area_map.unsqueeze(0).unsqueeze(0), (pad, pad, pad, pad), mode="replicate"
                )
                .squeeze(0)
                .squeeze(0)
            )
            new_distortion_ij = torch.nn.functional.pad(
                new_distortion_ij.unsqueeze(1),
                (pad, pad, pad, pad),
                mode="replicate",
            ).squeeze(1)
            new_distortion_IJ = torch.nn.functional.pad(
                new_distortion_IJ.unsqueeze(1),
                (pad, pad, pad, pad),
                mode="replicate",
            ).squeeze(1)
        kwargs = {
            "pixel_area_map": new_area_map,
            "sipA": self.sipA,
            "sipB": self.sipB,
            "sipAP": self.sipAP,
            "sipBP": self.sipBP,
            "distortion_ij": new_distortion_ij,
            "distortion_IJ": new_distortion_IJ,
            "_data": torch.zeros(
                (self.data.shape[0] * upsample + 2 * pad, self.data.shape[1] * upsample + 2 * pad),
                dtype=self.data.dtype,
                device=self.data.device,
            ),
            "CD": self.CD.value / upsample,
            "crpix": (self.crpix + 0.5) * upsample + pad - 0.5,
            "crtan": self.crtan.value,
            "crval": self.crval.value,
            "zeropoint": self.zeropoint,
            "identity": self.identity,
            "name": self.name + "_model",
            **kwargs,
        }
        return SIPModelImage(**kwargs)



from .model_object import ComponentModel

__all__ = ["SkyModel"]


class SkyModel(ComponentModel):
    """prototype class for any sky background model. This simply imposes
    that the center is a locked parameter, not involved in the
    fit. Also, a sky model object has no psf mode or integration mode
    by default since it is intended to change over much larger spatial
    scales than the psf or pixel size.

    """

    _model_type = "sky"
    usable = False

    def initialize(self):
        """Initialize the sky model, this is called after the model is
        created and before it is used. This is where we can set the
        center to be a locked parameter.
        """
        if not self.center.initialized:
            target_area = self.target[self.window]
            self.center.value = target_area.center
        super().initialize()
        self.center.to_static()

    @property
    def psf_mode(self):
        return "none"

    @psf_mode.setter
    def psf_mode(self, val):
        pass

    @property
    def integrate_mode(self):
        return "none"

    @integrate_mode.setter
    def integrate_mode(self, val):
        pass



import torch
from caskade import forward

from .base import Model
from ..image import ModelImage, PSFImage
from ..errors import InvalidTarget
from .mixins import SampleMixin


__all__ = ["PSFModel"]


class PSFModel(SampleMixin, Model):
    """Prototype point source (typically a star) model, to be subclassed
    by other point source models which define specific behavior.

    PSF_Models behave differently than component models. For starters,
    their target image must be a PSF_Image object instead of a
    Target_Image object. PSF_Models also don't define a "center"
    variable since their center is always (0,0) just like a
    PSF_Image. A PSF_Model will never be convolved with a PSF_Model
    (that's it's job!), so a lot of the sampling method is simpler.

    """

    _parameter_specs = {
        "center": {"units": "arcsec", "value": (0.0, 0.0), "shape": (2,)},
    }
    _model_type = "psf"
    usable = False

    # The sampled PSF will be normalized to a total flux of 1 within the window
    normalize_psf = True

    # Parameters which are treated specially by the model object and should not be updated directly when initializing
    _options = ("normalize_psf",)

    def initialize(self):
        pass

    @forward
    def transform_coordinates(self, x, y, center):
        return x - center[0], y - center[1]

    # Fit loop functions
    ######################################################################
    @forward
    def sample(self, window=None):
        """Evaluate the model on the space covered by an image object. This
        function properly calls integration methods. This should not
        be overloaded except in special cases.

        This function is designed to compute the model on a given
        image or within a specified window. It takes care of sub-pixel
        sampling, recursive integration for high curvature regions,
        and proper alignment of the computed model with the original
        pixel grid. The final model is then added to the requested
        image.

        Args:
          image (Optional[Image]): An AstroPhot Image object (likely a Model_Image)
                                     on which to evaluate the model values. If not
                                     provided, a new Model_Image object will be created.
          window (Optional[Window]): A window within which to evaluate the model.
                                   Should only be used if a subset of the full image
                                   is needed. If not provided, the entire image will
                                   be used.

        Returns:
          Image: The image with the computed model values.

        """
        # Create an image to store pixel samples
        working_image = self.target[self.window].model_image()
        working_image.data = self.sample_image(working_image)

        # normalize to total flux 1
        if self.normalize_psf:
            working_image.normalize()

        return working_image

    def fit_mask(self):
        return torch.zeros_like(self.target[self.window].mask, dtype=torch.bool)

    @property
    def target(self):
        try:
            return self._target
        except AttributeError:
            return None

    @target.setter
    def target(self, target):
        if target is None:
            self._target = None
        elif not isinstance(target, PSFImage):
            raise InvalidTarget(f"Target for PSF_Model must be a PSF_Image, not {type(target)}")
        try:
            del self._target  # Remove old target if it exists
        except AttributeError:
            pass

        self._target = target

    @forward
    def __call__(self, window=None) -> ModelImage:
        return self.sample(window=window)



from .image_object import Image, ImageList
from ..errors import InvalidImage

__all__ = ["ModelImage", "ModelImageList"]


######################################################################
class ModelImage(Image):
    """Image object which represents the sampling of a model at the given
    coordinates of the image. Extra arithmetic operations are
    available which can update model values in the image. The whole
    model can be shifted by less than a pixel to account for sub-pixel
    accuracy.

    """

    def fluxdensity_to_flux(self):
        self._data = self.data * self.pixel_area


######################################################################
class ModelImageList(ImageList):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not all(isinstance(image, (ModelImage, ModelImageList)) for image in self.images):
            raise InvalidImage(
                f"Model_Image_List can only hold Model_Image objects, not {tuple(type(image) for image in self.images)}"
            )



# =============================================================================
# Fit a single model to a target image
#
# This is a quick script to fit a single model to a target image. You should set
# the parameters under PARAMETERS to be appropriate for your data. The script
# will load the target image, mask, psf, and variance image (if available) and
# fit the model to the target image. The script will save the model image,
# residual image, and covariance matrix to the current directory. This script is
# intended for quick easy fits, users more comfortable with configuration file
# style behavior, and as a starting point to build a more complex analysis.
#
# Run this script with:
# >>> python single_model_fit.py
# =============================================================================

import astrophot as ap
import numpy as np
from astropy.io import fits
import matplotlib.pyplot as plt

# PARAMETERS
######################################################################
name = "object_name"  # used for saving files
target_file = "<required>.fits"  # can be a numpy array instead
psf_file = None  # "<path to psf>.fits" # can be a numpy array instead
zeropoint = 22.5  # mag
initial_params = None  # e.g. {"center": [3, 3], "q": {"value": 0.8, "locked": True}}
window = None  # None to fit whole image, otherwise (xmin,xmax,ymin,ymax) pixels
initial_sky = None  # If None, sky will be estimated
sky_locked = False
model_type = "sersic galaxy model"
# Extra parameters
######################################################################
save_model_image = True
save_residual_image = True
save_covariance_matrix = True
target_hdu = 0  # FITS file index for image data
psf_hdu = 0
sky_model_type = "flat sky model"
######################################################################

# load target
# ---------------------------------------------------------------------
print("loading target")
target = ap.TargetImage(
    filename=target_file,
    hduext=target_hdu,
    zeropoint=zeropoint,
)

# PSF
if isinstance(psf_file, str):
    print("loading psf")
    hdu = fits.open(psf_file)
    psf_data = np.array(hdu[psf_hdu].data, dtype=np.float64)
    target.psf = target.psf_image(data=psf_data)
elif psf_file is None:
    psf = None
else:
    target.psf = target.psf_image(data=psf_file)

# Create Model
# ---------------------------------------------------------------------
model_object = ap.Model(
    name=name,
    model_type=model_type,
    target=target,
    psf_convolve=True if psf_file is not None else False,
    **initial_params,
    window=window,
)
model_sky = ap.Model(
    name="sky",
    model_type=sky_model_type,
    target=target,
    I=initial_sky if initial_sky is not None else {},
    window=window,
)
if sky_locked:
    model_sky.to_static()
model = ap.Model(
    name="astrophot model",
    model_type="group model",
    target=target,
    models=[model_sky, model_object],
)

# Fit the model
# ---------------------------------------------------------------------
print("Initializing model")
model.initialize()
print("Fitting model")
result = ap.fit.LM(model, verbose=1).fit()

# Report Results
# ----------------------------------------------------------------------
print(model)
totmag = model_object.total_magnitude().detach().cpu().numpy()
totmag_err = model_object.total_magnitude_uncertainty().detach().cpu().numpy()
print(f"Total Magnitude: {totmag} +- {totmag_err}")

model.save_state(f"{name}_parameters.hdf5")
if save_model_image:
    model().save(f"{name}_model_image.fits")
    fig, ax = plt.subplots()
    ap.plots.model_image(fig, ax, model)
    plt.savefig(f"{name}_model_image.jpg")
    plt.close()
    if hasattr(model_object, "radial_model"):
        fig, ax = plt.subplots(figsize=(8, 8))
        ap.plots.radial_light_profile(fig, ax, model_object)
        plt.savefig(f"{name}_radial_light_profile.jpg")
        plt.close()
if save_residual_image:
    (target - model()).save(f"{name}_residual_image.fits")
    fig, ax = plt.subplots()
    ap.plots.residual_image(fig, ax, model, normalize_residuals=True)
    plt.savefig(f"{name}_residual_image.jpg")
    plt.close()

if save_covariance_matrix:
    np.save(f"{name}_covariance_matrix.npy", result.covariance_matrix.detach().cpu().numpy())
    fig, ax = ap.plots.covariance_matrix(
        result.covariance_matrix.detach().cpu().numpy(),
        model.parameters.vector_values().detach().cpu().numpy(),
        model.parameters.vector_names(),
    )
    fig.suptitle("Parameter Covariance")
    plt.savefig(f"{name}_covariance_matrix.pdf")
    plt.close()



import torch

from ..utils.decorators import ignore_numpy_warnings
from .psf_model_object import PSFModel
from .mixins import RadialMixin
from ..param import forward

__all__ = ("AiryPSF",)


class AiryPSF(RadialMixin, PSFModel):
    """The Airy disk is an analytic description of the diffraction pattern
    for a circular aperture.

    The diffraction pattern is described exactly by the configuration
    of the lens system under the assumption that all elements are
    perfect. This expression goes as:

    .. math::

      I(\\theta) = I_0\\left[\\frac{2J_1(x)}{x}\\right]^2

      x = ka\\sin(\\theta) = \\frac{2\\pi a r}{\\lambda R}

    where :math:`I(\\theta)` is the intensity as a function of the
    angular position within the diffraction system along its main
    axis, :math:`I_0` is the central intensity of the airy disk,
    :math:`J_1` is the Bessel function of the first kind of order one,
    :math:`k = \\frac{2\\pi}{\\lambda}` is the wavenumber of the
    light, :math:`a` is the aperture radius, :math:`r` is the radial
    position from the center of the pattern, :math:`R` is the distance
    from the circular aperture to the observation plane.

    In the `Airy_PSF` class we combine the parameters
    :math:`a,R,\\lambda` into a single ratio to be optimized (or fixed
    by the optical configuration).

    """

    _model_type = "airy"
    _parameter_specs = {
        "I0": {"units": "flux/arcsec^2", "value": 1.0, "shape": ()},
        "aRL": {"units": "a/(R lambda)", "shape": ()},
    }
    usable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    def initialize(self):
        super().initialize()

        if self.I0.initialized and self.aRL.initialized:
            return
        icenter = self.target.plane_to_pixel(*self.center.value)

        if not self.I0.initialized:
            mid_chunk = self.target.data[
                int(icenter[0]) - 2 : int(icenter[0]) + 2,
                int(icenter[1]) - 2 : int(icenter[1]) + 2,
            ]
            self.I0.dynamic_value = torch.mean(mid_chunk) / self.target.pixel_area
        if not self.aRL.initialized:
            self.aRL.dynamic_value = (5.0 / 8.0) * 2 * self.target.pixelscale

    @forward
    def radial_model(self, R, I0, aRL):
        x = 2 * torch.pi * aRL * R
        return I0 * (2 * torch.special.bessel_j1(x) / x) ** 2



import numpy as np

from matplotlib.patches import Ellipse
from matplotlib import pyplot as plt
from scipy.stats import norm
from .visuals import main_pallet

__all__ = ("covariance_matrix",)


def covariance_matrix(
    covariance_matrix,
    mean,
    labels=None,
    figsize=(10, 10),
    reference_values=None,
    ellipse_colors=main_pallet["primary1"],
    showticks=True,
    **kwargs,
):
    num_params = covariance_matrix.shape[0]
    fig, axes = plt.subplots(num_params, num_params, figsize=figsize)
    plt.subplots_adjust(wspace=0.0, hspace=0.0)

    for i in range(num_params):
        for j in range(num_params):
            ax = axes[i, j]

            if i == j:
                x = np.linspace(
                    mean[i] - 3 * np.sqrt(covariance_matrix[i, i]),
                    mean[i] + 3 * np.sqrt(covariance_matrix[i, i]),
                    100,
                )
                y = norm.pdf(x, mean[i], np.sqrt(covariance_matrix[i, i]))
                ax.plot(x, y, color=ellipse_colors, lw=1.5)
                ax.set_xlim(
                    mean[i] - 3 * np.sqrt(covariance_matrix[i, i]),
                    mean[i] + 3 * np.sqrt(covariance_matrix[i, i]),
                )
                if reference_values is not None:
                    ax.axvline(reference_values[i], color=main_pallet["pop"], linestyle="-", lw=1)
            elif j < i:
                cov = covariance_matrix[np.ix_([j, i], [j, i])]
                lambda_, v = np.linalg.eig(cov)
                lambda_ = np.sqrt(lambda_)
                angle = np.rad2deg(np.arctan2(v[1, 0], v[0, 0]))
                for k in [1, 2]:
                    ellipse = Ellipse(
                        xy=(mean[j], mean[i]),
                        width=lambda_[0] * k * 2,
                        height=lambda_[1] * k * 2,
                        angle=angle,
                        edgecolor=ellipse_colors,
                        facecolor="none",
                        lw=1.5,
                    )
                    ax.add_artist(ellipse)

                # Set axis limits
                margin = 3
                ax.set_xlim(
                    mean[j] - margin * np.sqrt(covariance_matrix[j, j]),
                    mean[j] + margin * np.sqrt(covariance_matrix[j, j]),
                )
                ax.set_ylim(
                    mean[i] - margin * np.sqrt(covariance_matrix[i, i]),
                    mean[i] + margin * np.sqrt(covariance_matrix[i, i]),
                )

                if reference_values is not None:
                    ax.axvline(reference_values[j], color=main_pallet["pop"], linestyle="-", lw=1)
                    ax.axhline(reference_values[i], color=main_pallet["pop"], linestyle="-", lw=1)

            if j > i:
                ax.axis("off")

            if i < num_params - 1:
                ax.set_xticklabels([])
            else:
                if labels is not None:
                    ax.set_xlabel(labels[j])
            if not showticks:
                ax.yaxis.set_major_locator(plt.NullLocator())

            if j > 0:
                ax.set_yticklabels([])
            else:
                if labels is not None:
                    ax.set_ylabel(labels[i])
            if not showticks:
                ax.xaxis.set_major_locator(plt.NullLocator())

    return fig, ax


if __name__ == "__main__":

    fig, ax = covariance_matrix(np.array([[4, -2], [-2, 4]]), np.array([0, 0]))
    plt.show()
